@@ -17,6 +17,10 @@
 
 #include "Rmpi.h"
 
+#ifndef Win32
+#include <dlfcn.h>
+#endif
+
 static MPI_Comm	*comm;
 static MPI_Status *status;
 static MPI_Datatype *datatype;
@@ -29,10 +33,28 @@ static int REQUEST_MAXSIZE=10000;
 SEXP mpi_initialize(){
 	int i,flag;
 	MPI_Initialized(&flag);
-	if (flag)
-		return AsInt(1);
-	else {	
-		MPI_Init((void *)0,(void *)0);
+
+#ifndef MPI2
+        static int fake_argc = 1;
+       	char *fake_argv[1];
+        char *fake_argv0 = "R";
+#endif
+
+if (flag)
+	return AsInt(1);
+	else {
+	
+#ifdef OPENMPI
+	dlopen("libmpi.so.0", RTLD_GLOBAL);
+#endif
+
+#ifndef MPI2
+   	fake_argv[0] = (char *)&fake_argv0;
+       	MPI_Init(&fake_argc, (char ***)(void*)&fake_argv);
+#else
+	MPI_Init((void *)0,(void *)0);
+#endif
+
 		MPI_Errhandler_set(MPI_COMM_WORLD, MPI_ERRORS_RETURN);
 		MPI_Errhandler_set(MPI_COMM_SELF, MPI_ERRORS_RETURN);
 		comm=(MPI_Comm *)Calloc(COMM_MAXSIZE, MPI_Comm); 
@@ -66,7 +88,7 @@ SEXP mpi_get_processor_name (){
     PROTECT (sexp_name  = allocVector (STRSXP, 1));
 	name = (char *)Calloc(MPI_MAX_PROCESSOR_NAME, char);
 	MPI_Get_processor_name(name, &resultlen);
-	SET_STRING_ELT(sexp_name, 0, COPY_TO_USER_STRING(name));
+	SET_STRING_ELT(sexp_name, 0, mkChar(name));
 	UNPROTECT(1);
 	Free(name);
 
@@ -110,17 +132,23 @@ SEXP mpi_info_create(SEXP sexp_info){
 }
 
 SEXP mpi_info_set(SEXP sexp_info, SEXP sexp_key, SEXP sexp_value){
-	return AsInt(erreturn(mpi_errhandler(MPI_Info_set(info[INTEGER(sexp_info)[0]], 
-		CHAR (STRING_ELT (sexp_key,0)),
-		CHAR (STRING_ELT (sexp_value,0))))));
+	return AsInt(erreturn(mpi_errhandler(MPI_Info_set(info[INTEGER(sexp_info)[0]],
+		CHAR2( STRING_ELT (sexp_key,0)), CHAR2(STRING_ELT(sexp_value,0))))));
 }
 
-SEXP mpi_info_get(SEXP sexp_info, SEXP sexp_key, SEXP sexp_valuelen, SEXP sexp_value){
+SEXP mpi_info_get(SEXP sexp_info, SEXP sexp_key, SEXP sexp_valuelen){
 	int flag;
+        char *value;
+	SEXP sexp_value;
+
+    	PROTECT (sexp_value  = allocVector (STRSXP, 1));
+ 	value = (char *)Calloc(INTEGER(sexp_valuelen)[0], char);  
 	mpi_errhandler(MPI_Info_get(info[INTEGER(sexp_info)[0]], 
-		CHAR (STRING_ELT (sexp_key,0)),
-		INTEGER(sexp_valuelen)[0],
-		CHAR (STRING_ELT (sexp_value,0)), &flag));
+		CHAR2( STRING_ELT (sexp_key,0)), 
+		INTEGER(sexp_valuelen)[0], value, &flag));
+        SET_STRING_ELT(sexp_value, 0, mkChar(value));
+        UNPROTECT(1); 
+	Free(value);
 	return sexp_value;
 }
 
@@ -180,42 +208,49 @@ SEXP mpi_gather(SEXP sexp_sdata,
 				   SEXP sexp_rdata,
 				   SEXP sexp_root,
 				   SEXP sexp_comm){
-	int len=LENGTH(sexp_sdata), commn=INTEGER(sexp_comm)[0], root=INTEGER(sexp_root)[0];
-	
+	int len, rlen, commn=INTEGER(sexp_comm)[0], root=INTEGER(sexp_root)[0];
+	char *rdata;
+	SEXP sexp_rdata2 = NULL;
+
 	switch (INTEGER(sexp_type)[0]){
 	case 1:
+ 		len=LENGTH(sexp_sdata);
 		mpi_errhandler(MPI_Gather(INTEGER(sexp_sdata), len, MPI_INT, 
 			INTEGER(sexp_rdata), len, MPI_INT, root, comm[commn]));
 		break;
 	case 2:
+ 		len=LENGTH(sexp_sdata);
 		mpi_errhandler(MPI_Gather(REAL(sexp_sdata), len, MPI_DOUBLE, 
 			REAL(sexp_rdata), len, MPI_DOUBLE, root, comm[commn]));
 		break;
-/*	case 3:
-		MPI_Comm_rank(comm[commn], &myrank);
-		if (myrank==rank) {
-			for (i=0; i<len; i++){
-				slen=strlen(CHAR (STRING_ELT ((sexp_data),i)));    
-				MPI_Bcast(CHAR (STRING_ELT ((sexp_data),i)), slen, MPI_CHAR, rank, comm[commn]);
-			}
-		}
-		else {
-			slen=strlen(CHAR(STRING_ELT(sexp_data,0)));
-			tmp=(char *)R_alloc(slen, sizeof(char));
-			for (i=0; i < len; i++){
-				MPI_Bcast(tmp,slen,MPI_CHAR,rank, comm[commn]);
-				SET_STRING_ELT(sexp_data, i, COPY_TO_USER_STRING(tmp));
-			}
-		}
+	case 3: 
+		len=LENGTH(STRING_ELT(sexp_sdata,0));
+		rlen=LENGTH(STRING_ELT(sexp_rdata,0));
+
+        	PROTECT (sexp_rdata2  = allocVector (STRSXP, 1));
+        	rdata = (char *)Calloc(rlen, char);
+                MPI_Gather(CHAR2 (STRING_ELT ((sexp_sdata),0)), len, MPI_CHAR,
+                        rdata, len, MPI_CHAR, root, comm[commn]);
+        	SET_STRING_ELT(sexp_rdata2, 0, mkChar(rdata));
+        	UNPROTECT(1);
+		Free(rdata);
 		break;
+	case 4:
+                len=LENGTH(sexp_sdata);
+                mpi_errhandler(MPI_Gather(RAW(sexp_sdata), len, MPI_BYTE,
+                        RAW(sexp_rdata), len, MPI_BYTE, root, comm[commn]));
+                break;
 	default:
-		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
-		mpi_errhandler(MPI_Bcast(REAL(sexp_data), 1, datatype[0], rank, comm[commn]));
+		PROTECT(sexp_sdata=AS_NUMERIC(sexp_sdata));
+		mpi_errhandler(MPI_Bcast(REAL(sexp_sdata), 1, datatype[0], root, comm[commn]));
 		UNPROTECT(1);
-		break;		*/
+		break;		
 	}
 
-	return sexp_rdata;
+	if (INTEGER(sexp_type)[0]==3)
+		return sexp_rdata2;
+	else
+		return sexp_rdata;
 }
 
 SEXP mpi_gatherv(SEXP sexp_sdata,
@@ -224,8 +259,10 @@ SEXP mpi_gatherv(SEXP sexp_sdata,
 				   SEXP sexp_recvcounts,
 				   SEXP sexp_root,
 				   SEXP sexp_comm){
-	int len=LENGTH(sexp_sdata), commn=INTEGER(sexp_comm)[0], root=INTEGER(sexp_root)[0];
-	int *displs, gsize, rank, i;
+	int len, rlen, commn=INTEGER(sexp_comm)[0], root=INTEGER(sexp_root)[0];
+	int *displs=NULL, gsize, rank, i;
+	char *rdata;
+	SEXP sexp_rdata2 = NULL;
 	
 	MPI_Comm_size(comm[commn], &gsize);
 	MPI_Comm_rank(comm[commn], &rank);
@@ -238,41 +275,50 @@ SEXP mpi_gatherv(SEXP sexp_sdata,
 
 	switch (INTEGER(sexp_type)[0]){
 	case 1:
+		len=LENGTH(sexp_sdata);
 		mpi_errhandler(MPI_Gatherv(INTEGER(sexp_sdata), len, MPI_INT, 
 			INTEGER(sexp_rdata), INTEGER(sexp_recvcounts), displs, MPI_INT, 
 				root, comm[commn]));
 		break;
 	case 2:
+		len=LENGTH(sexp_sdata);
 		mpi_errhandler(MPI_Gatherv(REAL(sexp_sdata), len, MPI_DOUBLE, 
 			REAL(sexp_rdata), INTEGER(sexp_recvcounts), displs, 
 				MPI_DOUBLE, root, comm[commn]));
 		break;
-/*	case 3:
-		MPI_Comm_rank(comm[commn], &myrank);
-		if (myrank==rank) {
-			for (i=0; i<len; i++){
-				slen=strlen(CHAR (STRING_ELT ((sexp_data),i)));    
-				MPI_Bcast(CHAR (STRING_ELT ((sexp_data),i)), slen, MPI_CHAR, rank, comm[commn]);
-			}
-		}
-		else {
-			slen=strlen(CHAR(STRING_ELT(sexp_data,0)));
-			tmp=(char *)R_alloc(slen, sizeof(char));
-			for (i=0; i < len; i++){
-				MPI_Bcast(tmp,slen,MPI_CHAR,rank, comm[commn]);
-				SET_STRING_ELT(sexp_data, i, COPY_TO_USER_STRING(tmp));
-			}
-		}
+	case 3:
+		len=LENGTH(STRING_ELT(sexp_sdata,0));	
+		rlen=LENGTH(STRING_ELT(sexp_rdata,0));
+        
+                PROTECT (sexp_rdata2  = allocVector (STRSXP, 1));
+                rdata = (char *)Calloc(rlen, char);
+                MPI_Gatherv(CHAR2 (STRING_ELT ((sexp_sdata),0)),len,MPI_CHAR,
+		     rdata, INTEGER(sexp_recvcounts),
+		     displs, MPI_CHAR, root, comm[commn]);
+                SET_STRING_ELT(sexp_rdata2, 0, mkChar(rdata));
+                UNPROTECT(1);
+		Free(rdata);
 		break;
+	case 4:
+                len=LENGTH(sexp_sdata);
+                mpi_errhandler(MPI_Gatherv(RAW(sexp_sdata), len, MPI_BYTE,
+                        RAW(sexp_rdata), INTEGER(sexp_recvcounts), displs,
+                                MPI_BYTE, root, comm[commn]));
+                break;
+
 	default:
-		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
-		mpi_errhandler(MPI_Bcast(REAL(sexp_data), 1, datatype[0], rank, comm[commn]));
+		PROTECT(sexp_sdata=AS_NUMERIC(sexp_sdata));
+		mpi_errhandler(MPI_Bcast(REAL(sexp_sdata), 1, datatype[0], rank, comm[commn]));
 		UNPROTECT(1);
-		break;		*/
+		break;	
 	}
 	if (rank == root)
 		Free(displs);
-	return sexp_rdata;
+
+	if (INTEGER(sexp_type)[0]==3)
+		return sexp_rdata2;
+	else
+		return sexp_rdata;
 }
 
 SEXP mpi_scatter(SEXP sexp_sdata,
@@ -280,42 +326,50 @@ SEXP mpi_scatter(SEXP sexp_sdata,
 				   SEXP sexp_rdata,
 				   SEXP sexp_root,
 				   SEXP sexp_comm){
-	int len=LENGTH(sexp_rdata), commn=INTEGER(sexp_comm)[0], root=INTEGER(sexp_root)[0];
-	
+	int 	len, rlen;
+	int	commn=INTEGER(sexp_comm)[0], root=INTEGER(sexp_root)[0];
+	char 	*rdata;
+	SEXP 	sexp_rdata2 = NULL;
+
 	switch (INTEGER(sexp_type)[0]){
 	case 1:
+ 		len=LENGTH(sexp_rdata);
 		mpi_errhandler(MPI_Scatter(INTEGER(sexp_sdata), len, MPI_INT, 
 			INTEGER(sexp_rdata), len, MPI_INT, root, comm[commn]));
 		break;
 	case 2:
+		len=LENGTH(sexp_rdata);
 		mpi_errhandler(MPI_Scatter(REAL(sexp_sdata), len, MPI_DOUBLE, 
 			REAL(sexp_rdata), len, MPI_DOUBLE, root, comm[commn]));
 		break;
-/*	case 3:
-		MPI_Comm_rank(comm[commn], &myrank);
-		if (myrank==rank) {
-			for (i=0; i<len; i++){
-				slen=strlen(CHAR (STRING_ELT ((sexp_data),i)));    
-				MPI_Bcast(CHAR (STRING_ELT ((sexp_data),i)), slen, MPI_CHAR, rank, comm[commn]);
-			}
-		}
-		else {
-			slen=strlen(CHAR(STRING_ELT(sexp_data,0)));
-			tmp=(char *)R_alloc(slen, sizeof(char));
-			for (i=0; i < len; i++){
-				MPI_Bcast(tmp,slen,MPI_CHAR,rank, comm[commn]);
-				SET_STRING_ELT(sexp_data, i, COPY_TO_USER_STRING(tmp));
-			}
-		}
-		break;
-	default:
-		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
-		mpi_errhandler(MPI_Bcast(REAL(sexp_data), 1, datatype[0], rank, comm[commn]));
-		UNPROTECT(1);
-		break;		*/
-	}
+	case 3:
+ 		len=LENGTH(STRING_ELT(sexp_rdata,0));
+                rlen=LENGTH(STRING_ELT(sexp_rdata,0));
 
-	return sexp_rdata;
+                PROTECT (sexp_rdata2  = allocVector (STRSXP, 1));
+                rdata = (char *)Calloc(rlen, char);
+		MPI_Scatter(CHAR2(STRING_ELT ((sexp_sdata),0)), len, MPI_CHAR,
+                       rdata, len, MPI_CHAR, root, comm[commn]);
+                SET_STRING_ELT(sexp_rdata2, 0, mkChar(rdata));
+                UNPROTECT(1);
+		Free(rdata);
+		break;
+	case 4:
+                len=LENGTH(sexp_rdata);
+                mpi_errhandler(MPI_Scatter(RAW(sexp_sdata), len, MPI_BYTE,
+                        RAW(sexp_rdata), len, MPI_BYTE, root, comm[commn]));
+                break;
+
+	default:
+		PROTECT(sexp_sdata=AS_NUMERIC(sexp_sdata));
+		mpi_errhandler(MPI_Bcast(REAL(sexp_sdata), 1, datatype[0], root, comm[commn]));
+		UNPROTECT(1);
+		break;		
+	}
+        if (INTEGER(sexp_type)[0]==3)
+                return sexp_rdata2;
+        else
+                return sexp_rdata;
 }
 
 SEXP mpi_scatterv(SEXP sexp_sdata,
@@ -324,9 +378,11 @@ SEXP mpi_scatterv(SEXP sexp_sdata,
 				  SEXP sexp_rdata,
 				  SEXP sexp_root,
 				  SEXP sexp_comm){
-	int len=LENGTH(sexp_rdata), commn=INTEGER(sexp_comm)[0], root=INTEGER(sexp_root)[0];
-	int *displs, gsize, rank, i;
-	
+	int len, rlen, commn=INTEGER(sexp_comm)[0], root=INTEGER(sexp_root)[0];
+	int gsize,rank,i,*displs=NULL;
+    	char *rdata;
+	SEXP sexp_rdata2 = NULL;
+
 	MPI_Comm_size(comm[commn], &gsize);
 	MPI_Comm_rank(comm[commn], &rank);
 	if (rank==root){
@@ -335,87 +391,104 @@ SEXP mpi_scatterv(SEXP sexp_sdata,
 		for (i=1; i < gsize; i++)
 			displs[i]=displs[i-1]+INTEGER(sexp_sendcounts)[i-1];
 	}
-
 	
 	switch (INTEGER(sexp_type)[0]){
 	case 1:
+		 len=LENGTH(sexp_rdata);
 		mpi_errhandler(MPI_Scatterv(INTEGER(sexp_sdata), INTEGER(sexp_sendcounts),
 			displs, MPI_INT, INTEGER(sexp_rdata), len, MPI_INT, 
 				root, comm[commn]));
 		break;
 	case 2:
+		 len=LENGTH(sexp_rdata);
 		mpi_errhandler(MPI_Scatterv(REAL(sexp_sdata), INTEGER(sexp_sendcounts),
 			displs, MPI_DOUBLE, REAL(sexp_rdata), len,  
 				MPI_DOUBLE, root, comm[commn]));
 		break;
-/*	case 3:
-		MPI_Comm_rank(comm[commn], &myrank);
-		if (myrank==rank) {
-			for (i=0; i<len; i++){
-				slen=strlen(CHAR (STRING_ELT ((sexp_data),i)));    
-				MPI_Bcast(CHAR (STRING_ELT ((sexp_data),i)), slen, MPI_CHAR, rank, comm[commn]);
-			}
-		}
-		else {
-			slen=strlen(CHAR(STRING_ELT(sexp_data,0)));
-			tmp=(char *)R_alloc(slen, sizeof(char));
-			for (i=0; i < len; i++){
-				MPI_Bcast(tmp,slen,MPI_CHAR,rank, comm[commn]);
-				SET_STRING_ELT(sexp_data, i, COPY_TO_USER_STRING(tmp));
-			}
-		}
+	case 3:
+                len=LENGTH(STRING_ELT(sexp_rdata,0));
+                rlen=LENGTH(STRING_ELT(sexp_rdata,0));
+
+                PROTECT (sexp_rdata2  = allocVector (STRSXP, 1));
+                // rdata = (char *)R_alloc(rlen, sizeof(char));
+                rdata = (char *)Calloc(rlen, char);
+                MPI_Scatterv(CHAR2 (STRING_ELT ((sexp_sdata),0)), INTEGER(sexp_sendcounts),displs, 
+			MPI_CHAR,rdata, len, MPI_CHAR, root, comm[commn]);
+                SET_STRING_ELT(sexp_rdata2, 0, mkChar(rdata));
+                UNPROTECT(1);
+		Free(rdata);
 		break;
+	case 4:
+                len=LENGTH(sexp_rdata);
+                mpi_errhandler(MPI_Scatterv(RAW(sexp_sdata), INTEGER(sexp_sendcounts),
+                        displs, MPI_BYTE, RAW(sexp_rdata), len,
+                                MPI_BYTE, root, comm[commn]));
+                break;
+
 	default:
-		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
-		mpi_errhandler(MPI_Bcast(REAL(sexp_data), 1, datatype[0], rank, comm[commn]));
+		PROTECT(sexp_sdata=AS_NUMERIC(sexp_sdata));
+		mpi_errhandler(MPI_Bcast(REAL(sexp_sdata), 1, datatype[0], rank, comm[commn]));
 		UNPROTECT(1);
-		break;		*/
+		break;		
 	}
 	if (rank == root)
+
 		Free(displs);
-	return sexp_rdata;
+
+        if (INTEGER(sexp_type)[0]==3)
+                return sexp_rdata2;
+        else
+                return sexp_rdata;
 }
 
 SEXP mpi_allgather(SEXP sexp_sdata,
 				   SEXP sexp_type,
 				   SEXP sexp_rdata,
 				   SEXP sexp_comm){
-	int len=LENGTH(sexp_sdata), commn=INTEGER(sexp_comm)[0];
+	int len, rlen, commn=INTEGER(sexp_comm)[0];
+	char *rdata;
+	SEXP sexp_rdata2 = NULL;
 	
 	switch (INTEGER(sexp_type)[0]){
 	case 1:
+		len=LENGTH(sexp_sdata);
 		mpi_errhandler(MPI_Allgather(INTEGER(sexp_sdata), len, MPI_INT, 
 			INTEGER(sexp_rdata), len, MPI_INT, comm[commn]));
 		break;
 	case 2:
+		len=LENGTH(sexp_sdata);
 		mpi_errhandler(MPI_Allgather(REAL(sexp_sdata), len, MPI_DOUBLE, 
 			REAL(sexp_rdata), len, MPI_DOUBLE, comm[commn]));
 		break;
-/*	case 3:
-		MPI_Comm_rank(comm[commn], &myrank);
-		if (myrank==rank) {
-			for (i=0; i<len; i++){
-				slen=strlen(CHAR (STRING_ELT ((sexp_data),i)));    
-				MPI_Bcast(CHAR (STRING_ELT ((sexp_data),i)), slen, MPI_CHAR, rank, comm[commn]);
-			}
-		}
-		else {
-			slen=strlen(CHAR(STRING_ELT(sexp_data,0)));
-			tmp=(char *)R_alloc(slen, sizeof(char));
-			for (i=0; i < len; i++){
-				MPI_Bcast(tmp,slen,MPI_CHAR,rank, comm[commn]);
-				SET_STRING_ELT(sexp_data, i, COPY_TO_USER_STRING(tmp));
-			}
-		}
+	case 3:
+	 	len=LENGTH(STRING_ELT(sexp_sdata,0));
+                rlen=LENGTH(STRING_ELT(sexp_rdata,0));
+
+                PROTECT (sexp_rdata2  = allocVector (STRSXP, 1));
+                rdata = (char *)Calloc(rlen, char);
+                MPI_Allgather(CHAR2 (STRING_ELT ((sexp_sdata),0)),len,
+			MPI_CHAR,rdata, len, MPI_CHAR, comm[commn]);
+                SET_STRING_ELT(sexp_rdata2, 0, mkChar(rdata));
+                UNPROTECT(1);
+		Free(rdata);
 		break;
+ 	case 4:
+                len=LENGTH(sexp_sdata);
+                mpi_errhandler(MPI_Allgather(RAW(sexp_sdata), len, MPI_BYTE,
+                        RAW(sexp_rdata), len, MPI_BYTE, comm[commn]));
+                break;
+
 	default:
-		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
-		mpi_errhandler(MPI_Bcast(REAL(sexp_data), 1, datatype[0], rank, comm[commn]));
+		PROTECT(sexp_sdata=AS_NUMERIC(sexp_sdata));
+		mpi_errhandler(MPI_Bcast(REAL(sexp_sdata), 1, datatype[0], 0, comm[commn]));
 		UNPROTECT(1);
-		break;		*/
+		break;
 	}
 
-	return sexp_rdata;
+        if (INTEGER(sexp_type)[0]==3)
+                return sexp_rdata2;
+        else
+                return sexp_rdata;
 }
 
 SEXP mpi_allgatherv(SEXP sexp_sdata,
@@ -423,7 +496,9 @@ SEXP mpi_allgatherv(SEXP sexp_sdata,
 				   SEXP sexp_rdata,
 				   SEXP sexp_recvcounts,
 				   SEXP sexp_comm){
-	int len=LENGTH(sexp_sdata), commn=INTEGER(sexp_comm)[0], *displs, gsize, i;
+	int len, rlen, commn=INTEGER(sexp_comm)[0], *displs, gsize, i;
+	char *rdata;
+	SEXP sexp_rdata2 = NULL;
 	
 	MPI_Comm_size(comm[commn], &gsize);
 	displs=(int *)Calloc(gsize, int);
@@ -433,40 +508,47 @@ SEXP mpi_allgatherv(SEXP sexp_sdata,
 
 	switch (INTEGER(sexp_type)[0]){
 	case 1:
-		mpi_errhandler(MPI_Allgatherv(INTEGER(sexp_sdata), len, MPI_INT, 
-			INTEGER(sexp_rdata), INTEGER(sexp_recvcounts), displs, MPI_INT, 
-				comm[commn]));
+		len=LENGTH(sexp_sdata);
+		mpi_errhandler(MPI_Allgatherv(INTEGER(sexp_sdata), len, MPI_INT,
+			INTEGER(sexp_rdata), INTEGER(sexp_recvcounts), displs,
+				 MPI_INT,comm[commn]));
 		break;
 	case 2:
-		mpi_errhandler(MPI_Allgatherv(REAL(sexp_sdata), len, MPI_DOUBLE, 
+		len=LENGTH(sexp_sdata);
+		mpi_errhandler(MPI_Allgatherv(REAL(sexp_sdata), len, MPI_DOUBLE,
 			REAL(sexp_rdata), INTEGER(sexp_recvcounts), displs, 
 				MPI_DOUBLE, comm[commn]));
 		break;
-/*	case 3:
-		MPI_Comm_rank(comm[commn], &myrank);
-		if (myrank==rank) {
-			for (i=0; i<len; i++){
-				slen=strlen(CHAR (STRING_ELT ((sexp_data),i)));    
-				MPI_Bcast(CHAR (STRING_ELT ((sexp_data),i)), slen, MPI_CHAR, rank, comm[commn]);
-			}
-		}
-		else {
-			slen=strlen(CHAR(STRING_ELT(sexp_data,0)));
-			tmp=(char *)R_alloc(slen, sizeof(char));
-			for (i=0; i < len; i++){
-				MPI_Bcast(tmp,slen,MPI_CHAR,rank, comm[commn]);
-				SET_STRING_ELT(sexp_data, i, COPY_TO_USER_STRING(tmp));
-			}
-		}
+	case 3:
+	 	len=LENGTH(STRING_ELT(sexp_sdata,0));
+                rlen=LENGTH(STRING_ELT(sexp_rdata,0));
+
+                PROTECT (sexp_rdata2  = allocVector (STRSXP, 1));
+                rdata = (char *)Calloc(rlen, char);
+                MPI_Allgatherv(CHAR2 (STRING_ELT ((sexp_sdata),0)),len, MPI_CHAR, rdata,
+		      INTEGER(sexp_recvcounts), displs, MPI_CHAR, comm[commn]);
+                SET_STRING_ELT(sexp_rdata2, 0, mkChar(rdata));
+                UNPROTECT(1);
+		Free(rdata);
 		break;
+ 	case 4:
+                len=LENGTH(sexp_sdata);
+                mpi_errhandler(MPI_Allgatherv(RAW(sexp_sdata), len, MPI_BYTE,
+                        RAW(sexp_rdata), INTEGER(sexp_recvcounts), displs,
+                                MPI_BYTE, comm[commn]));
+                break;
+
 	default:
-		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
-		mpi_errhandler(MPI_Bcast(REAL(sexp_data), 1, datatype[0], rank, comm[commn]));
+		PROTECT(sexp_sdata=AS_NUMERIC(sexp_sdata));
+		mpi_errhandler(MPI_Bcast(REAL(sexp_sdata), 1, datatype[0], 0, comm[commn]));
 		UNPROTECT(1);
-		break;		*/
+		break;	
 	}
 	Free(displs);
-	return sexp_rdata;
+        if (INTEGER(sexp_type)[0]==3)
+                return sexp_rdata2;
+        else
+                return sexp_rdata;
 }
 
 SEXP mpi_bcast(SEXP sexp_data,
@@ -474,9 +556,11 @@ SEXP mpi_bcast(SEXP sexp_data,
 			   SEXP	sexp_rank,
 			   SEXP sexp_comm){
 
-	int i, len=LENGTH(sexp_data), type=INTEGER(sexp_type)[0];
-	int rank=INTEGER(sexp_rank)[0], commn=INTEGER(sexp_comm)[0],slen;
+	int len=LENGTH(sexp_data), type=INTEGER(sexp_type)[0];
+	int rank=INTEGER(sexp_rank)[0], root,  commn=INTEGER(sexp_comm)[0],slen;
 	int errcode=0;
+	char *rdata;
+	SEXP sexp_data2 = NULL;
 
 	switch (type){
 	case 1:
@@ -486,11 +570,24 @@ SEXP mpi_bcast(SEXP sexp_data,
 		mpi_errhandler(MPI_Bcast(REAL(sexp_data), len, MPI_DOUBLE, rank, comm[commn]));
 		break;
 	case 3:
-		for (i=0; i<len; i++){
-			slen=LENGTH(STRING_ELT ((sexp_data),i)); 
-			MPI_Bcast(CHAR (STRING_ELT ((sexp_data),i)), slen, MPI_CHAR, rank, comm[commn]);
+        	MPI_Comm_rank(comm[commn], &root);
+		slen=LENGTH(STRING_ELT (sexp_data,0)); 
+		if (rank==root) 
+			MPI_Bcast(CHAR2 (STRING_ELT (sexp_data,0)), slen, 
+				MPI_CHAR, rank, comm[commn]);
+		else {
+                	PROTECT (sexp_data2  = allocVector (STRSXP, 1));
+	               	rdata = (char *)Calloc(slen, char);
+                       	MPI_Bcast(rdata, slen, MPI_CHAR, rank, comm[commn]);
+			SET_STRING_ELT(sexp_data2, 0, mkChar(rdata));
+			UNPROTECT(1);
+			Free(rdata);
 		}
 		break;
+	case 4:
+                errcode=MPI_Bcast(RAW(sexp_data), len, MPI_BYTE, rank, comm[commn]);
+                break;
+
 	default:
 		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
 		mpi_errhandler(MPI_Bcast(REAL(sexp_data), 1, datatype[0], rank, comm[commn]));
@@ -504,8 +601,12 @@ SEXP mpi_bcast(SEXP sexp_data,
 		Rprintf("%s\n",errmsg);
 		return mkString("error");
 	}
-	else
-		return sexp_data;
+	else {
+        	if ((INTEGER(sexp_type)[0]==3) && (rank!=root))
+                	return sexp_data2;
+        	else
+                	return sexp_data;
+	}
 }
 
 SEXP mpi_send(SEXP sexp_data, 
@@ -513,7 +614,7 @@ SEXP mpi_send(SEXP sexp_data,
 			  SEXP sexp_dest, 
 			  SEXP sexp_tag,
 			  SEXP sexp_comm){
-	int i,slen,len=LENGTH(sexp_data),type=INTEGER(sexp_type)[0], dest=INTEGER(sexp_dest)[0];
+	int slen,len=LENGTH(sexp_data),type=INTEGER(sexp_type)[0], dest=INTEGER(sexp_dest)[0];
 	int commn=INTEGER(sexp_comm)[0], tag=INTEGER(sexp_tag)[0];
 
 	switch (type){
@@ -524,11 +625,13 @@ SEXP mpi_send(SEXP sexp_data,
 		mpi_errhandler(MPI_Send(REAL(sexp_data), len, MPI_DOUBLE, dest, tag, comm[commn]));
 		break;
 	case 3:
-		for (i=0; i<len; i++){
-			slen=LENGTH(STRING_ELT(sexp_data,i));
-			MPI_Send(CHAR(STRING_ELT(sexp_data,i)),slen, MPI_CHAR, dest, tag, comm[commn]); 
-		}
+		slen=LENGTH(STRING_ELT(sexp_data,0));
+		MPI_Send(CHAR2(STRING_ELT(sexp_data,0)),slen, MPI_CHAR, dest, tag, comm[commn]); 
 		break;
+        case 4:
+                MPI_Send(RAW(sexp_data),len, MPI_BYTE, dest, tag, comm[commn]);                
+                break;
+
 	default:
 		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
 		mpi_errhandler(MPI_Send(REAL(sexp_data), 1, datatype[0], dest, tag, comm[commn]));
@@ -544,9 +647,11 @@ SEXP mpi_recv(SEXP sexp_data,
 			  SEXP sexp_tag,
 			  SEXP sexp_comm,
 			  SEXP sexp_status){
-	int i, len=LENGTH(sexp_data), type=INTEGER(sexp_type)[0], source=INTEGER(sexp_source)[0];
+	int len=LENGTH(sexp_data), type=INTEGER(sexp_type)[0], source=INTEGER(sexp_source)[0];
 	int tag=INTEGER(sexp_tag)[0],commn=INTEGER(sexp_comm)[0], statusn=INTEGER(sexp_status)[0];
 	int slen;
+	char *rdata;
+	SEXP sexp_data2 = NULL;
 
 	switch (type){
 	case 1:
@@ -559,11 +664,18 @@ SEXP mpi_recv(SEXP sexp_data,
 		break;
 	case 3:
 		slen=LENGTH(STRING_ELT(sexp_data,0));
-		for (i=0; i < len; i++){
-			MPI_Recv(CHAR(STRING_ELT(sexp_data,i)),
-				slen,MPI_CHAR,source,tag, comm[commn],&status[statusn]);
-		}
+                PROTECT (sexp_data2  = allocVector (STRSXP, 1));
+                rdata = (char *)Calloc(slen, char);
+		MPI_Recv(rdata, slen,MPI_CHAR,source,tag, comm[commn],&status[statusn]);
+                SET_STRING_ELT(sexp_data2, 0, mkChar(rdata));
+                UNPROTECT(1);
+		Free(rdata);
 		break;
+	case 4:          
+		mpi_errhandler(MPI_Recv(RAW(sexp_data), len, MPI_BYTE, source, tag, comm[commn],
+                        &status[statusn]));
+                break;
+
 	default:
 		PROTECT(sexp_data=AS_NUMERIC(sexp_data));
 		mpi_errhandler(MPI_Recv(REAL(sexp_data), 1, datatype[0], source, tag, comm[commn],
@@ -571,7 +683,10 @@ SEXP mpi_recv(SEXP sexp_data,
 		UNPROTECT(1);
 		break;		
 	}
-	return sexp_data;
+        if (INTEGER(sexp_type)[0]==3)
+                return sexp_data2;
+        else
+		return sexp_data;
 }
 
 SEXP mpi_reduce(SEXP sexp_send, 
@@ -581,8 +696,8 @@ SEXP mpi_reduce(SEXP sexp_send,
 				SEXP sexp_comm){
 	int len=LENGTH(sexp_send), type=INTEGER(sexp_type)[0], dest=INTEGER(sexp_dest)[0];
 	int commn=INTEGER(sexp_comm)[0], intop = INTEGER(sexp_op)[0];
-	MPI_Op op;
-	SEXP sexp_recv;
+	MPI_Op op= MPI_SUM;
+	SEXP sexp_recv = NULL;
 
 	switch(intop){
 	case 1:
@@ -667,8 +782,8 @@ SEXP mpi_allreduce(SEXP sexp_send,
 				   SEXP sexp_comm){
 	int len=LENGTH(sexp_send), type=INTEGER(sexp_type)[0], commn=INTEGER(sexp_comm)[0];
 	int intop = INTEGER(sexp_op)[0];
-	MPI_Op op;
-	SEXP sexp_recv;
+	MPI_Op op = MPI_SUM;
+	SEXP sexp_recv = NULL;
 
 	switch(intop){
 	case 1:
@@ -763,7 +878,7 @@ SEXP mpi_probe(SEXP sexp_source, SEXP sexp_tag, SEXP sexp_comm, SEXP sexp_status
 
 SEXP mpi_get_count(SEXP sexp_status, SEXP sexp_type){
 	SEXP sexp_count;
-	MPI_Datatype datatype;
+	MPI_Datatype datatype = MPI_DATATYPE_NULL;
 	
 	switch(INTEGER(sexp_type)[0]){
 	case 1:
@@ -775,6 +890,10 @@ SEXP mpi_get_count(SEXP sexp_status, SEXP sexp_type){
 	case 3:
 		datatype=MPI_CHAR;
 		break;
+ 	case 4:
+                datatype=MPI_BYTE;
+                break;
+
 	}
 	
 	PROTECT (sexp_count = allocVector(INTSXP, 1));
@@ -864,15 +983,15 @@ SEXP mpi_comm_spawn (SEXP sexp_slave,
 
 	slaverrcode = (int *)Calloc(nslave, int);
 	if (len==0)
-		mpi_errhandler(MPI_Comm_spawn (CHAR (STRING_ELT (sexp_slave, 0)), MPI_ARGV_NULL, nslave,   
+		mpi_errhandler(MPI_Comm_spawn (CHAR2 (STRING_ELT (sexp_slave, 0)), MPI_ARGV_NULL, nslave,   
 					info[infon], root, MPI_COMM_SELF, &comm[intercommn],
 					slaverrcode)); 
 	else {
 		char **argv = (char **) R_alloc (len+1, sizeof (char *));
 		for (i = 0; i < len; i++)
-			argv[i] = CHAR (STRING_ELT (sexp_argv, i));
+			argv[i] = CHAR2 (STRING_ELT (sexp_argv, i));
 		argv[len] = NULL;
-		mpi_errhandler(MPI_Comm_spawn (CHAR (STRING_ELT (sexp_slave, 0)), argv, nslave,   
+		mpi_errhandler(MPI_Comm_spawn (CHAR2 (STRING_ELT (sexp_slave, 0)), argv, nslave,   
 					info[infon], root, MPI_COMM_SELF, &comm[intercommn],
 					slaverrcode)); 
 	}
@@ -896,7 +1015,7 @@ SEXP mpi_is_master(){
 	MPI_Comm master;
 	MPI_Comm_get_parent(&master);
 	check=(master==MPI_COMM_NULL);
-	MPI_Comm_free(&master);
+	if (!check) MPI_Comm_free(&master);
 	return AsInt(check);
 }
 
@@ -934,6 +1053,8 @@ SEXP mpi_sendrecv(SEXP sexp_senddata,
     int recvcount=LENGTH(sexp_recvdata), recvtype=INTEGER(sexp_recvtype)[0];
     int source=INTEGER(sexp_source)[0], recvtag=INTEGER(sexp_recvtag)[0];
     int commn=INTEGER(sexp_comm)[0],statusn=INTEGER(sexp_status)[0];
+    char *rdata;
+    SEXP sexp_recvdata2 = NULL;
 
     switch(sendtype){
         case 1:
@@ -952,12 +1073,20 @@ SEXP mpi_sendrecv(SEXP sexp_senddata,
                     break;
                 case 3:
                     rlen=LENGTH(STRING_ELT(sexp_recvdata,0)); 
-                    MPI_Sendrecv(INTEGER(sexp_senddata),sendcount,
-			MPI_INT, dest, sendtag, 
-			CHAR(STRING_ELT(sexp_recvdata,0)), rlen, 
-			MPI_CHAR, source, recvtag, comm[commn],  
-			&status[statusn]);
+                    PROTECT (sexp_recvdata2  = allocVector (STRSXP, 1));
+                    rdata = (char *)Calloc(rlen, char);
+                    MPI_Sendrecv(INTEGER(sexp_senddata),sendcount, MPI_INT, dest, sendtag, 
+			rdata, rlen, MPI_CHAR, source, recvtag, comm[commn],  &status[statusn]);
+                    SET_STRING_ELT(sexp_recvdata2, 0, mkChar(rdata));
+                    UNPROTECT(1);
+                    Free(rdata);
                     break;
+ 		case 4:
+                     MPI_Sendrecv(INTEGER(sexp_senddata),sendcount,
+                        MPI_INT,dest,sendtag, RAW(sexp_recvdata),
+                        recvcount, MPI_BYTE, source,recvtag,
+                        comm[commn], &status[statusn]);
+		    break;
                 }
 	    break;
         case 2:
@@ -976,12 +1105,21 @@ SEXP mpi_sendrecv(SEXP sexp_senddata,
                     break;
                 case 3:
                     rlen=LENGTH(STRING_ELT(sexp_recvdata,0)); 
-                    MPI_Sendrecv(REAL(sexp_recvdata),sendcount,
-			MPI_DOUBLE, dest, sendtag, 
-			CHAR(STRING_ELT(sexp_recvdata,0)), rlen, 
-			MPI_CHAR, source, recvtag, comm[commn],  
-			&status[statusn]);
+                    PROTECT (sexp_recvdata2  = allocVector (STRSXP, 1));
+                    rdata = (char *)Calloc(rlen, char);
+                    MPI_Sendrecv(REAL(sexp_senddata),sendcount, MPI_DOUBLE, dest, sendtag, 
+			rdata, rlen, MPI_CHAR, source, recvtag, comm[commn],  &status[statusn]);
+                    SET_STRING_ELT(sexp_recvdata2, 0, mkChar(rdata));
+                    UNPROTECT(1);
+                    Free(rdata);
                     break;
+ 		case 4:
+                    MPI_Sendrecv(REAL(sexp_senddata),sendcount,
+                        MPI_DOUBLE,dest,sendtag, RAW(sexp_recvdata),
+                        recvcount, MPI_BYTE, source,recvtag,
+                        comm[commn], &status[statusn]);
+                    break;
+
                 }
             break;
         case 3: 
@@ -989,30 +1127,75 @@ SEXP mpi_sendrecv(SEXP sexp_senddata,
   
             switch(recvtype){
                	case 1:
-                    MPI_Sendrecv(CHAR(STRING_ELT(sexp_senddata,0)),slen, 
+                    MPI_Sendrecv(CHAR2(STRING_ELT(sexp_senddata,0)),slen, 
 			MPI_CHAR,dest,sendtag, INTEGER(sexp_recvdata), 
 			recvcount, MPI_INT, source,recvtag,
                         comm[commn], &status[statusn]);
                     break;
                 case 2:                       
-		    MPI_Sendrecv(CHAR(STRING_ELT(sexp_senddata,0)),slen, 
+		    MPI_Sendrecv(CHAR2(STRING_ELT(sexp_senddata,0)),slen, 
 			MPI_CHAR,dest,sendtag, REAL(sexp_recvdata), 
 			recvcount, MPI_DOUBLE, source,recvtag,
                         comm[commn], &status[statusn]);
                     break;
                 case 3:
                     rlen=LENGTH(STRING_ELT(sexp_recvdata,0));
-                        
-		    MPI_Sendrecv(CHAR(STRING_ELT(sexp_senddata,0)),slen,
-			MPI_CHAR, dest, sendtag, 
-			CHAR(STRING_ELT(sexp_recvdata,0)), rlen, 
-			MPI_CHAR, source, recvtag, comm[commn], 
-			&status[statusn]);
+                    PROTECT (sexp_recvdata2  = allocVector (STRSXP, 1));
+                    rdata = (char *)Calloc(rlen, char);
+
+		    MPI_Sendrecv(CHAR2(STRING_ELT(sexp_senddata,0)),slen, MPI_CHAR, dest, sendtag, 
+			rdata, rlen, MPI_CHAR, source, recvtag, comm[commn], &status[statusn]);
+                    SET_STRING_ELT(sexp_recvdata2, 0, mkChar(rdata));
+                    UNPROTECT(1);
+                    Free(rdata);
                     break;
+ 		case 4:
+                    MPI_Sendrecv(CHAR2(STRING_ELT(sexp_senddata,0)),slen,
+                        MPI_CHAR,dest,sendtag, RAW(sexp_recvdata),
+                        recvcount, MPI_BYTE, source,recvtag,
+                        comm[commn], &status[statusn]);
+                    break;
+
           	}
    		break;
+ 	case 4:
+            switch(recvtype){
+                case 1:
+                    MPI_Sendrecv(RAW(sexp_senddata),sendcount,
+                        MPI_BYTE,dest,sendtag, INTEGER(sexp_recvdata),
+                        recvcount, MPI_INT, source,recvtag,
+                        comm[commn], &status[statusn]);
+                    break;
+                case 2:
+                    MPI_Sendrecv(RAW(sexp_senddata),sendcount,
+                        MPI_BYTE,dest,sendtag, REAL(sexp_recvdata),
+                        recvcount, MPI_DOUBLE, source,recvtag,
+                        comm[commn], &status[statusn]);
+                    break;
+                case 3:
+                    rlen=LENGTH(STRING_ELT(sexp_recvdata,0));
+                    PROTECT (sexp_recvdata2  = allocVector (STRSXP, 1));
+                    rdata = (char *)Calloc(rlen, char);
+                    MPI_Sendrecv(RAW(sexp_senddata),sendcount, MPI_BYTE, dest, sendtag, 
+			rdata, rlen, MPI_CHAR, source, recvtag, comm[commn],  &status[statusn]);
+                    SET_STRING_ELT(sexp_recvdata2, 0, mkChar(rdata));
+                    UNPROTECT(1);
+                    Free(rdata);
+                    break;
+                case 4:
+                    MPI_Sendrecv(RAW(sexp_senddata),sendcount,
+                        MPI_BYTE,dest,sendtag, RAW(sexp_recvdata),
+                        recvcount, MPI_BYTE, source,recvtag,
+                        comm[commn], &status[statusn]);
+                    break;
+                }
+            break;
+
 	    }
-    return sexp_recvdata;          
+    if (recvtype==3)
+	return sexp_recvdata2;
+    else
+    	return sexp_recvdata;          
 }
 
 SEXP mpi_sendrecv_replace(SEXP sexp_data,
@@ -1024,11 +1207,13 @@ SEXP mpi_sendrecv_replace(SEXP sexp_data,
         SEXP sexp_comm, 
         SEXP sexp_status)
 {
-        int i, slen;
+        int slen;
         int len=LENGTH(sexp_data), type=INTEGER(sexp_type)[0];
         int dest=INTEGER(sexp_dest)[0], sendtag=INTEGER(sexp_sendtag)[0];
         int source=INTEGER(sexp_source)[0],recvtag=INTEGER(sexp_recvtag)[0];
         int commn=INTEGER(sexp_comm)[0],statusn=INTEGER(sexp_status)[0];
+	char *srdata;
+	SEXP sexp_data2 = NULL;
 
         switch (type){
         case 1:
@@ -1042,12 +1227,19 @@ SEXP mpi_sendrecv_replace(SEXP sexp_data,
 
         case 3:
                 slen=LENGTH(STRING_ELT(sexp_data,0));
-                for (i=0; i < len; i++){
-                        MPI_Sendrecv_replace(CHAR(STRING_ELT(sexp_data,i)),
-                                slen,MPI_CHAR, dest, sendtag, source, recvtag, 
-				comm[commn], &status[statusn]);  
-                }
+		PROTECT (sexp_data2  = allocVector (STRSXP, 1));
+		srdata= (char *)Calloc(slen, char);
+		strcpy(srdata, CHAR(STRING_ELT(sexp_data,0)));
+                MPI_Sendrecv_replace(srdata, slen,MPI_CHAR, dest, sendtag, source, recvtag, 
+				comm[commn], &status[statusn]); 
+		UNPROTECT(1);
+		Free(srdata); 
                 break;
+ 	case 4:
+                MPI_Sendrecv_replace(RAW(sexp_data), len, MPI_BYTE, dest,
+                sendtag, source, recvtag, comm[commn], &status[statusn]);
+                break;
+
         default:
                 PROTECT(sexp_data=AS_NUMERIC(sexp_data));
                 MPI_Sendrecv_replace(REAL(sexp_data), 1, datatype[0], dest, 
@@ -1055,7 +1247,10 @@ SEXP mpi_sendrecv_replace(SEXP sexp_data,
                         &status[statusn]);                        
                 break;
           }
-          return sexp_data;
+	if (type==3)
+		return sexp_data2;
+	else
+          	return sexp_data;
 }
 
 /************ cart dim *************************************/
@@ -1140,7 +1335,7 @@ SEXP mpi_isend(SEXP sexp_data,
                           SEXP sexp_tag,
                           SEXP sexp_comm,
                           SEXP sexp_request){
-        int i,slen,len=LENGTH(sexp_data),type=INTEGER(sexp_type)[0], dest=INTEGER(sexp_dest)[0];
+        int slen,len=LENGTH(sexp_data),type=INTEGER(sexp_type)[0], dest=INTEGER(sexp_dest)[0];
         int commn=INTEGER(sexp_comm)[0], tag=INTEGER(sexp_tag)[0], requestn=INTEGER(sexp_request)[0];
         
         switch (type){
@@ -1154,12 +1349,16 @@ SEXP mpi_isend(SEXP sexp_data,
 			&request[requestn]));
                 break;
         case 3:
-                for (i=0; i<len; i++){
-                        slen=LENGTH(STRING_ELT(sexp_data,i));
-                        mpi_errhandler(MPI_Isend(CHAR(STRING_ELT(sexp_data,i)),slen, MPI_CHAR, dest, 
+                slen=LENGTH(STRING_ELT(sexp_data,0));
+                mpi_errhandler(MPI_Isend(CHAR2(STRING_ELT(sexp_data,0)),slen, MPI_CHAR, dest, 
 			tag, comm[commn], &request[requestn]));
-                }
                 break;
+	case 4:
+
+                mpi_errhandler(MPI_Isend(RAW(sexp_data), len, MPI_BYTE, dest, tag, comm[commn],
+                        &request[requestn]));
+                break;
+
         default:
                 PROTECT(sexp_data=AS_NUMERIC(sexp_data));
                 mpi_errhandler(MPI_Isend(REAL(sexp_data), 1, datatype[0], dest, tag, comm[commn], 
@@ -1176,7 +1375,7 @@ SEXP mpi_irecv(SEXP sexp_data,
                           SEXP sexp_tag,
                           SEXP sexp_comm,
                           SEXP sexp_request){
-        int i,slen,len=LENGTH(sexp_data),type=INTEGER(sexp_type)[0], source=INTEGER(sexp_source)[0];
+        int slen,len=LENGTH(sexp_data),type=INTEGER(sexp_type)[0], source=INTEGER(sexp_source)[0];
         int commn=INTEGER(sexp_comm)[0], tag=INTEGER(sexp_tag)[0], requestn=INTEGER(sexp_request)[0];
 
         switch (type){
@@ -1189,12 +1388,15 @@ SEXP mpi_irecv(SEXP sexp_data,
 			&request[requestn]));
                 break;
         case 3:
-                for (i=0; i<len; i++){
-                        slen=LENGTH(STRING_ELT(sexp_data,i));
-                        mpi_errhandler(MPI_Irecv(CHAR(STRING_ELT(sexp_data,i)),slen, MPI_CHAR, 
+                slen=LENGTH(STRING_ELT(sexp_data,0));
+                        mpi_errhandler(MPI_Irecv(CHAR2(STRING_ELT(sexp_data,0)),slen, MPI_CHAR, 
 				source, tag, comm[commn], &request[requestn]));
-                }
                 break;
+ 	case 4:
+                mpi_errhandler(MPI_Irecv(RAW(sexp_data), len, MPI_BYTE, source, tag, comm[commn],
+                        &request[requestn]));
+                break;
+
         default:
                 PROTECT(sexp_data=AS_NUMERIC(sexp_data));
                 mpi_errhandler(MPI_Irecv(REAL(sexp_data), 1, datatype[0], source, tag, comm[commn], 
@@ -1237,7 +1439,7 @@ SEXP mpi_waitany(SEXP sexp_count, SEXP sexp_status){
 }
 
 SEXP mpi_testany(SEXP sexp_count, SEXP sexp_status){
-        int countn=INTEGER(sexp_count)[0], index, statusn=INTEGER(sexp_status)[0];
+        int countn=INTEGER(sexp_count)[0],  statusn=INTEGER(sexp_status)[0];
 		SEXP indexflag;
 		PROTECT (indexflag = allocVector(INTSXP, 2));
         mpi_errhandler(MPI_Testany(countn, request, &INTEGER(indexflag)[0],
@@ -1284,4 +1486,4 @@ SEXP mpi_request_get_status(SEXP sexp_request,  SEXP sexp_status){
         mpi_errhandler(MPI_Request_get_status(request[requestn], &flag, &status[statusn]));
         return AsInt(flag);
 }
-*/
\ No newline at end of file
+*/
